#!/usr/bin/env python
"""Unit tests for aho-corasick keyword string searching.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)
"""

import unittest

## import sys
## sys.path.insert(0, "build/lib.darwin-7.8.0-PowerMacintosh-2.3/")

import ahocorasick


class AhoCorasickTest(unittest.TestCase):
    def setUp(self):
        self.tree = ahocorasick.KeywordTree()


    def tearDown(self):
        self.tree = None


    def testKeywordAsPrefixOfAnother(self):
        """According to John, there's a problem with the matcher.
        this test case should expose the bug."""
        self.tree.add('foobar')
        self.tree.add('foo')
        self.tree.add('bar')
        self.tree.make()
        self.assertEqual((3, 6), self.tree.search('xxxfooyyy'))
        self.assertEqual((0, 3), self.tree.search('foo'))
        self.assertEqual((3, 6), self.tree.search('xxxbaryyy'))


    def testAnotherSearch(self):
        """Just to triangulate the search code.  We want to make sure
        that the implementation can do more than one search, at
        least."""
        self.tree.add("Python")
        self.tree.add("PLT Scheme")
        self.tree.make()
        self.assertEqual((19, 25),
                         self.tree.search("I am learning both Python and PLT Scheme"))
        self.assertEqual((0, 10),
                         self.tree.search("PLT Scheme is an interesting language."))


    def testSimpleConstruction(self):
        self.tree.add("foo")
        self.tree.add("bar")
        self.tree.make()
        self.assertEqual((10, 13),
                         self.tree.search("this is a foo message"))


    def testLongestSearch(self):
        self.tree.add("a")
        self.tree.add("alphabet");
        self.tree.make()
        self.assertEqual((0, 1), self.tree.search("alphabet soup"));
        self.assertEqual((0, 8), self.tree.search_long("alphabet soup"))
        self.assertEqual((13, 14), self.tree.search_long("yummy, I see an alphabet soup bowl"))
        


    def testSearchWithWholeMatch(self):
        """Make sure that longest search will match the whole string."""
        longString = "supercalifragilisticexpialidocious"
        self.tree.add(longString)
        self.tree.make()
        self.assertEqual((0, len(longString)),
                         self.tree.search(longString))


    def testSearchLongWithWholeMatch(self):
        """Make sure that longest search will match the whole string."""
        longString = "supercalifragilisticexpialidocious"
        self.tree.add(longString)
        self.tree.make()
        self.assertEqual((0, len(longString)),
                         self.tree.search_long(longString))

    def testSearchLongWithNoMatch(self):
        self.tree.add("foobar")
        self.tree.make()
        self.assertEqual(None, self.tree.search_long("fooba"))


    def testWithExpectedNonmatch(self):
        """Check to see that we don't always get a successful match."""
        self.tree.add("wise man")
        self.tree.make()
        self.assertEqual(
            None,
            self.tree.search("where fools and wise men fear to tread"))        


    def testEmptyConstruction(self):
        """Make sure that we can safely construct and dealloc a tree
        with no initial keywords.  Important because the C
        implementation assumes keywords exist on its dealloc, so we
        have to do some work on the back end to avoid silly segmentation
        errors."""
        tree = ahocorasick.KeywordTree()
        del tree


    def testEmptyMake(self):
        """Calling make() without adding any keywords should raise an
        AssertionError."""
        self.assertRaises(AssertionError, self.tree.make)


    def testDegenerateConstruction(self):
        """If we try searching before making, we should get an
        assertion error."""
        self.assertRaises(AssertionError,
                          lambda: self.tree.search("this is a test"))
    

    def testEmptyKeyword(self):
        """Trying to add an empty should raise an assertion error."""
        self.assertRaises(AssertionError,
                          lambda: self.tree.add(""))
        

    def testEmptyQuery(self):
        self.tree.add("hello world")
        self.tree.make()
        self.assertEqual(None, self.tree.search(""))
    

    def testStartposWithSearch(self):
        """Check to see if startpos does the right thing."""
        self.tree.add("wood")
        self.tree.add("woodchuck")
        self.tree.make()
        queryString = "howmuchwoodwouldawoodchuckchuck"
        self.assertEqual((7, 11), self.tree.search(queryString))
        self.assertEqual((17, 21), self.tree.search(queryString, startpos=11))
        self.assertEqual((17, 21), self.tree.search(queryString, 11))


    def testStartposWithSearchLong(self):
        """Check to see if startpos does the right thing."""
        self.tree.add("wood")
        self.tree.add("woodchuck")
        self.tree.make()
        queryString = "howmuchwoodwouldawoodchuckchuck"
        self.assertEqual((7, 11), self.tree.search_long(queryString))
        self.assertEqual((17, 26),
                         self.tree.search_long(queryString, startpos=11))
        self.assertEqual((17, 26), self.tree.search_long(queryString, 11))


    def testEmbeddedNulls(self):
        """Check to see if we can accept embedded nulls"""
        self.tree.add("hell\0 world")
        self.tree.make()
        self.assertEqual(None, self.tree.search("ello\0 world"))
        self.assertEqual((0, 11), self.tree.search("hell\0 world"))



    def testEmbeddedNullsAgain(self):
        self.tree.add("\0\0\0")
        self.tree.make()
        self.assertEqual((0, 3), self.tree.search("\0\0\0\0\0\0\0\0"))
        self.assertEqual((3, 6), self.tree.search("\0\0\0\0\0\0\0\0", 3))
        self.assertEqual(None, self.tree.search("\0\0\0\0\0\0\0\0", 6))
        

if __name__ == '__main__':
    unittest.main()
