/* Python wrapper code for Aho-Corasick code.  I know I should be
   using some wrapper-generating code, but I'm just practicing my C.
   *grin*

   Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

 */


#include <Python.h>
#include "structmember.h"
#include "aho-corasick.h"

/* Keyword Tree structure definition. */
typedef struct {
	PyObject_HEAD
	int count;
	int made;
	aho_corasick_t* tree;
} ahocorasick_KeywordTree;



/* Constructs a new KeywordTree. */
static PyObject* 
ahocorasick_KeywordTree_new(PyTypeObject *type,
			    PyObject *args,
			    PyObject *kwds) {
	ahocorasick_KeywordTree *self;
/* 	printf("In new\n");	/\* debug *\/ */
	self = (ahocorasick_KeywordTree *) type->tp_alloc(type, 0);
	if (self != NULL) {
		self->tree = PyMem_Malloc(sizeof(aho_corasick_t));
		if (self->tree == NULL) {
			Py_DECREF(self);
			return NULL;
		}
/* 		printf("Initializing tree.\n"); /\* debug *\/ */
		aho_corasick_init(self->tree);
		self->count = 0;
		self->made = 0;
	}
	return (PyObject*) self;
}


/* Initializes a KeywordTree. */
static int
ahocorasick_KeywordTree_init(ahocorasick_KeywordTree *self,
			     PyObject *args,
			     PyObject *kwargs) {
	
/* 	printf("In init\n");	/\* debug *\/ */
	/* fill me in */
	return 0;
}



/* Deallocates the memory used in allocating this thing. */
static void
ahocorasick_KeywordTree_dealloc(ahocorasick_KeywordTree *self) {
/* 	printf("In dealloc\n");	/\* debug *\/ */
	if (self->count > 0) {
		aho_corasick_destroy(self->tree);
	}
	PyMem_Free(self->tree);
	self->ob_type->tp_free((PyObject*) self);
}


/* Adds a new keyword to the KeywordTree. */
static PyObject*
ahocorasick_KeywordTree_add(ahocorasick_KeywordTree *self,
			    PyObject *args) {
	unsigned char *newKeyword;
	size_t n;
	if (! PyArg_ParseTuple(args, "s#", &newKeyword, &n)) {
		return NULL;
	}

	/* Check for empty string: the underlying C implementation function
	   aho_corasick_addstring() crashes on empty string input, so let's
	   catch that before we enter. */
	if (n == 0) {
		PyErr_SetString(PyExc_AssertionError,
				"add() cannot take the empty string");
		return NULL;
	}

	if (self->made) {
		PyErr_SetString(PyExc_AssertionError, 
				"add() cannot be called once a tree has been finalized with make()");
		return NULL;
	}

/* 	printf("Adding %s\n", newKeyword); /\* debug *\/ */

	aho_corasick_addstring(self->tree, newKeyword, n);
	self->count++;

	Py_INCREF(Py_None);
	return Py_None;
}




/* Given a string, searches for the first keyword.  Either returns None, or a
   2-tuple (start, end).  Since search() and search_long() are so similar, I
   extract the common elements of both here, and specialize by using a helper
   function pointer. */
static PyObject*
ahocorasick_KeywordTree_basesearch(ahocorasick_KeywordTree *self,
				   PyObject *args, PyObject *kwargs,
				   ahocorasick_KeywordTree_search_helper_t helper) {
	unsigned char *queryString;
	size_t start, end;
	static char *kwlist[] = {"query", "startpos", NULL};
	size_t startpos = 0;
	size_t n;		/* length of queryString */
	if (! PyArg_ParseTupleAndKeywords
	    (args, kwargs, "s#|i", kwlist, &queryString, &n, &startpos)) {
		return NULL;
	}

	/* Before searching, the KeywordTree must have been make()ed. earlier,
	   or else the search doesn't work. */
	if (!self->made) {
		PyErr_SetString(PyExc_AssertionError,
				"make() must be called before search() to finalize tree construction.");
		return NULL;
	}
	
	if ((*helper)(self->tree, 
		      queryString, n,
		      startpos,
		      &start, &end)) {
	  return Py_BuildValue("(ll)", start, end);
	}

	/* If we get to this point, the search has failed. */
	Py_INCREF(Py_None);
	return Py_None; 
}



/* Given a string, searches for the first keyword.  Either returns None, or a
   2-tuple (start, end). */
static PyObject*
ahocorasick_KeywordTree_search(ahocorasick_KeywordTree *self,
			       PyObject *args, PyObject *kwargs) {
	return ahocorasick_KeywordTree_basesearch
		(self, args, kwargs,
		 ahocorasick_KeywordTree_search_helper);
}



/* Given a string, searches for the first keyword.  Either returns None, or a
   2-tuple (start, end).  Tries to match as long a keyword as it can.*/
static PyObject*
ahocorasick_KeywordTree_search_long(ahocorasick_KeywordTree *self,
				    PyObject *args, PyObject *kwargs) {
	return ahocorasick_KeywordTree_basesearch
		(self, args, kwargs,
		 ahocorasick_KeywordTree_search_long_helper);
}






/* Once the keywords have been passed into the tree, maketree does some final
   construction of the keyword tree.

   We must make sure that maketree has not be called multiple times.  If it
   is, ignore the call.
*/
static PyObject*
ahocorasick_KeywordTree_maketree(ahocorasick_KeywordTree *self,
				 PyObject *args) {
	/* fixme: only call this if self->count is positive! */
	if (!self->made) {
		if (self->count == 0) {
			PyErr_SetString(PyExc_AssertionError,
					"make() can not be called until at least one string has been add()ed.");
			return NULL;
		}
		aho_corasick_maketree(self->tree);
		self->made = 1;
	}
	Py_INCREF(Py_None);
	return Py_None;
}




static PyMemberDef ahocorasick_KeywordTree_members[] = {
	{NULL}			/* sentinel */
};



static PyMethodDef ahocorasick_KeywordTree_methods[] = {
	/* The PyCFunction casts here are necessary to make the C compiler
	   happy.  These functions take ahocorasick_KeywordTree pointers as
	   their first argument. */
	{"add", (PyCFunction) ahocorasick_KeywordTree_add, METH_VARARGS,
	 "Add a new keyword to the KeywordTree." },
	{"search", (PyCFunction) ahocorasick_KeywordTree_search, METH_VARARGS | METH_KEYWORDS,
	 "Search for a keyword.  Either returns a 2-tuple \
(start, end), or None." },
	{"search_long", (PyCFunction) ahocorasick_KeywordTree_search_long, METH_VARARGS | METH_KEYWORDS,
	 "Search for a keyword.  Either returns a 2-tuple \
(start, end), or None.  Tries for longest match." },
	{"make", (PyCFunction) ahocorasick_KeywordTree_maketree, METH_VARARGS,
	 "Finishes KeywordTree construction." },
	{NULL}			/* sentinel */
};





/* Let's define the type table stuff. */
static PyTypeObject ahocorasick_KeywordTreeType = {
    PyObject_HEAD_INIT(NULL)
    0,                                                /*ob_size*/
    "ahocorasick.KeywordTree",                        /*tp_name*/
    sizeof(ahocorasick_KeywordTree),                  /*tp_basicsize*/
    0,                                                /*tp_itemsize*/
    (destructor)ahocorasick_KeywordTree_dealloc,      /*tp_dealloc*/
    0,                         /*tp_print*/
    0,                         /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,        /*tp_flags*/
    "KeywordTree objects",           /* tp_doc */
    0,                         /* tp_traverse */
    0,                         /* tp_clear */
    0,                         /* tp_richcompare */
    0,                         /* tp_weaklistoffset */
    0,                         /* tp_iter */
    0,                         /* tp_iternext */
    ahocorasick_KeywordTree_methods,             /* tp_methods */
    ahocorasick_KeywordTree_members,             /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc) ahocorasick_KeywordTree_init,      /* tp_init */
    0,                         /* tp_alloc */
    ahocorasick_KeywordTree_new,                 /* tp_new */
};





/* ********************************************************************** */

/* Module stuff below: */



/* Module method table */

static PyMethodDef AhoCorasickMethods[] = {
  {NULL, NULL, 0, NULL}		/* sentinel */
};




PyMODINIT_FUNC
initahocorasick(void) {
	PyObject* m;

	if (PyType_Ready(&ahocorasick_KeywordTreeType) < 0)
		return;

	m = Py_InitModule3("ahocorasick", AhoCorasickMethods,
			   "Aho-Corasick keyword tree");

	if (m == NULL)
		return;

	Py_INCREF(&ahocorasick_KeywordTreeType);
	PyModule_AddObject(m, "KeywordTree", 
			   (PyObject*)&ahocorasick_KeywordTreeType);
}

